/* Copyright (C) 1999 Chris Vine, G3XXF

This program is distributed under the General Public Licence, version 2.
For particulars of this and relevant disclaimers see the file
COPYRIGHT distributed with the source files.

*/

#include "screen.h"

termios Screen::startsettings;
int Screen::object_count = 0;
int Screen::have_colour = FALSE;
int Screen::x_cursor_pos;
int Screen::y_cursor_pos;
int Screen::nrows;
int Screen::ncols;
WinList Screen::active_popups;

Screen::Screen(void (*curses_init_ptr)(void)): winlist_guard(FALSE) {
    if (!object_count) {
        if ((void*)curses_init_ptr != 0) {
	    tcgetattr(fileno(stdin), &startsettings);
	    setupterm(0, fileno(stdout), (int*)0);
	    nrows = tigetnum("lines");
	    ncols = tigetnum("cols");
	    if (nrows < 15 || ncols < 80) {
	        cerr << "Terminal screen too small" << endl;
		exit(TERM_ERROR);
	    }
	    if (nrows > 80) nrows = 80;
	    if (ncols > 120) ncols = 120;

	    initscr();

	    if (has_colors() && start_color() == OK) {
	        have_colour = TRUE;
// set COLOR_PAIR(1) and COLOR_PAIR(2) and COLOR_PAIR(3) for general use as default colours
		init_pair(1, COLOR_WHITE, COLOR_BLUE);
		init_pair(2, COLOR_WHITE, COLOR_RED);
		init_pair(3, COLOR_YELLOW, COLOR_BLUE);
	    }
	    wclear(stdscr);
	    curses_init_ptr();
	    curs_set(1);
	}
	else {
	    cerr << "The first definition of a Screen object must pass a valid function pointer\n"
	            "to the constructor in order to set up the keyboard\n"
		    "Please examine header file screen.h for further details" << endl;
	    exit(TERM_ERROR);
	}
    }
    object_count++;
}

Screen::~Screen(void) {
    object_count--;
    if (!object_count) {
        endwin();
	tcsetattr(fileno(stdin), TCSANOW, &startsettings);
    }
}

void Screen::set_cursor_status(Cursor_status cursor) {
    if (cursor == Screen::on) {
        leaveok(win, FALSE);
	cursor_status = Screen::on;
    }
    else {
        leaveok(win, TRUE);
	cursor_status = Screen::off;
    }
}

void Screen::write(char letter) {
    waddch(win, (uchar)letter);
    if (cursor_status == on) {
        wnoutrefresh(win);
	get_cur_pos();
    }
}

void Screen::write(char letter, int rowpos, int colpos) {
    wmove(win, rowpos, colpos);
    waddch(win, (uchar)letter);
    if (cursor_status == on) {
        wnoutrefresh(win);
        get_cur_pos();
    }
}

void Screen::write(const char* message) {
    waddstr(win, message);
    if (cursor_status == on) {
        wnoutrefresh(win);
        get_cur_pos();
    }
}

void Screen::write(const char* message, int rowpos, int colpos) {
    wmove(win, rowpos, colpos);
    waddstr(win, message);
    if (cursor_status == on) {
        wnoutrefresh(win);
        get_cur_pos();
    }
}

void Screen::show(void) {
    touchwin(win);
    if (cursor_status == off) winrefresh();
    else {
// if switching beween windows, both of which show cursors when in foreground, we want to reseat the cursor
        wrefresh(win);
	get_cur_pos();
	if (!active_popups.is_empty() && !winlist_guard) {
	    winlist_guard = TRUE;
	    refresh_winlist(active_popups);
	    winlist_guard = FALSE;
	}
    }
}

void Screen::winrefresh(void) {
    wnoutrefresh(win);
    set_cur_pos();
    doupdate();
    if (!active_popups.is_empty() && !winlist_guard) {
        winlist_guard = TRUE;
	refresh_winlist(active_popups);
	winlist_guard = FALSE;
    }
}


MainScreen::MainScreen(void (*curses_init_ptr)(void), Tnc& tnc_): Screen(curses_init_ptr), tnc(tnc_) {
    win = stdscr;
    bottom_row = nrows - 1;
    if (have_colour) bkgd(COLOR_PAIR(1));
    int count;
    attron(A_REVERSE);
    move(0, 0);
    for(count = 0; count < ncols; count++) addch('-');
    refresh();
    mvaddstr(0, ncols/2 - 20, " G3XXF terminal program for KAM-Plus ");
    refresh();
    move(1, 0);
    for(count = 0; count < ncols; count++) addch(' ');
    mvaddstr(1, 0, "Call:              HF Mode:          Capture:          HF Send:");
    refresh();
    move(7, 0);
    for(count = 0; count < ncols; count++) addch('-');
    refresh();
    move(bottom_row, 0);
    for(count = 0; count < ncols; count++) addch(' ');
    refresh();
    mvaddstr(bottom_row, 13, "|  F1 for help   |");
    refresh();
    attroff(A_REVERSE);

    set_cursor_status(off);

    display_capture_status();
    display_send_mode_status();
    display_current_stream();
    display_connected_status();
    display_mode();
}


void MainScreen::display_capture_status(void){
    attron(A_REVERSE);
    move(1, 46);
    wnoutrefresh(stdscr);
    if (tnc.tnc_func.capturefile_flag == FALSE) addstr("Off    ");
    else {
        if (tnc.tnc_func.capture_stream.port) addstr(" HF-");
	else addstr("VHF-");
	if (!tnc.tnc_func.capture_stream.port || tnc.tnc_func.hfmode == Tnc_func::packet)  {
	    addch(' ');
	    addch((tnc.tnc_func.capture_stream.stream | 0x40) + 1);
	    addch(' ');
	}
	else addstr("Tor");
    }
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE);
}

void MainScreen::display_send_mode_status(void) {
    attron(A_REVERSE);
    move(1, 64);
    wnoutrefresh(stdscr);
    if (prog_func.send_mode == Prog_func::word) addstr("Word ");
    else if (prog_func.send_mode == Prog_func::guard) addstr("Guard");
    else addstr("Line ");
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE);
}

void MainScreen::display_callsign(void) {
    attron(A_REVERSE);
    move(1, 5);
    wnoutrefresh(stdscr);
    addstr("              ");
    wnoutrefresh(stdscr);
    move(1, 5);
    wnoutrefresh(stdscr);
    if (tnc.tnc_func.active_port && tnc.tnc_func.hfmode != Tnc_func::packet 
	&& tnc.tnc_func.hisCall_lock == Tnc_func::on) addch('*');
    else addch(' ');
    addstr(tnc.tnc_func.hisCall[tnc.tnc_func.active_stream()][tnc.tnc_func.active_port]);
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE);
}

void MainScreen::display_current_stream(void) {
    if (have_colour) attron(A_REVERSE | COLOR_PAIR(2));
    else attron(A_REVERSE | A_BOLD);
    move(1, ncols - 7);
    wnoutrefresh(stdscr);
    if (tnc.tnc_func.active_port) addstr(" HF-");
    else addstr("VHF-");
    if (!tnc.tnc_func.active_port || tnc.tnc_func.hfmode == Tnc_func::packet)  {
        addch(' ');
        addch((tnc.tnc_func.active_stream() | 0x40) + 1);
        addch(' ');
    }
    else addstr("Tor");
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE | A_BOLD | COLOR_PAIR(2));
}


void MainScreen::display_connected_status(void) {
    attron(A_REVERSE);
    move(bottom_row, 0);
    wnoutrefresh(stdscr);
    if (!tnc.tnc_func.stream_status[tnc.tnc_func.active_stream()][tnc.tnc_func.active_port]) {
        addstr("Disconnected");
    }
    else  addstr("Connected   ");
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE);
}

void MainScreen::display_freebytes(void) {
    attron(A_REVERSE);
    move(bottom_row, ncols - 4);
    wnoutrefresh(stdscr);
    printw("%4d", tnc.get_freebytes());
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE);
}

void MainScreen::display_mode(void) {
    attron(A_REVERSE);
    move(1, 28);
    wnoutrefresh(stdscr);
    switch(tnc.tnc_func.hfmode) {
    case Tnc_func::packet:
        addstr("Packet");
	break;
    case Tnc_func::pactor:
        addstr("Pactor");
	break;
    case Tnc_func::rtty:
        addstr("RTTY  ");
	break;
    case Tnc_func::ascii:
        addstr("ASCII ");
	break;
    case Tnc_func::amtor:
        addstr("Amtor ");
	break;
    case Tnc_func::lamtor:
        addstr("Lamtor");
	break;
    case Tnc_func::fec:
        addstr("FEC   ");
	break;
    case Tnc_func::gtor:
        addstr("Gtor  ");
	break;
    case Tnc_func::gmon:
        addstr("Gmon  ");
	break;
    case Tnc_func::tor:
        addstr("Tor   ");
	break;
    case Tnc_func::cw:
        addstr("CW    ");
	break;
    }
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE);
}

void MainScreen::display_time(void) {
    time_t time_val;
    time(&time_val);
    tm* time_ptr = gmtime(&time_val);
    attron(A_REVERSE);
    mvprintw(7, ncols/2 - 5, " %02d:%02d ", time_ptr->tm_hour, time_ptr->tm_min);
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE);
}

void MainScreen::make_torline(void) {
    attron(A_REVERSE);
    move(bottom_row, 31);
    if (tnc.tnc_func.active_port && tnc.tnc_func.hfmode != Tnc_func::packet) {
        addstr( "      | RX |      |     |     |     |      |");
    }
    else addstr("                                            ");
    wnoutrefresh(stdscr);
    setsyx(y_cursor_pos, x_cursor_pos);
    doupdate();
    attroff(A_REVERSE);
}

void MainScreen::update_torinfo(unsigned char info) { //info has a default value of 255
    if (info != 255) info_byte = info; // if info = 255 we just use the value stored in info_byte
    if (tnc.tnc_func.active_port && tnc.tnc_func.hfmode != Tnc_func::packet) {
        attron(A_REVERSE);
	move(bottom_row, 44);
        if (info_byte & 1) addstr("IDLE");
	else addstr("    ");
	move(bottom_row, 51);
	if (info_byte & 2) addstr("ERR");
	else if (info_byte & 4) addstr("CMB");
	else if (info_byte & 8) addstr("RQ ");
	else addstr("   ");
	move(bottom_row, 69);
	if (info_byte & 16) addstr("HUFF");
	else addstr("    ");
	move(bottom_row, 57);
	if (info_byte & 32) addstr("ISS");
	else if (tnc.tnc_func.stream_status[0][1] == Tnc_func::connected) addstr("IRS");
	else addstr("   ");
	move(bottom_row, 63);
	if (info_byte & 64) addstr("200");
	else if (info_byte & 128) addstr("300");
	else if (tnc.tnc_func.stream_status[0][1] == Tnc_func::connected) addstr("100");
	else addstr("   ");
	wnoutrefresh(stdscr);
	setsyx(y_cursor_pos, x_cursor_pos);
	doupdate();
	attroff(A_REVERSE);
    }
}

void MainScreen::update_txinfo(unsigned char tx) { //tx has a default value of 255
    if (tx != 255) tx_byte = tx;
    if (tnc.tnc_func.active_port && tnc.tnc_func.hfmode != Tnc_func::packet) {
        attron(A_REVERSE);
	move(bottom_row, 39);
	if (tx_byte & 2) {
	    if (have_colour) attron(COLOR_PAIR(2));
	    addstr("TX");
	}
	else addstr("RX");
	wnoutrefresh(stdscr);
	setsyx(y_cursor_pos, x_cursor_pos);
	doupdate();
	attroff(A_REVERSE | COLOR_PAIR(2));
    }
}

void MainScreen::update_lockinfo(void) {
    if (tnc.tnc_func.active_port && tnc.tnc_func.hfmode != Tnc_func::packet) {
        attron(A_REVERSE);
	move(bottom_row, 32);
	if (tnc.tnc_func.speed_lock == Tnc_func::on) addstr("LOCK");
	else addstr("    ");
	wnoutrefresh(stdscr);
	setsyx(y_cursor_pos, x_cursor_pos);
	doupdate();
	attroff(A_REVERSE);
    }
}

HelpWin::HelpWin(void): buffer_ptr(0) {
    win = newwin(nrows - 9, ncols, 8, 0);
    scrollok(win, TRUE);
    set_cursor_status(off);
    if (have_colour) wbkgd(win, COLOR_PAIR(1));

    char* filename = new char[strlen(prog_func.filedir) + 10];
    strcpy(filename, prog_func.filedir);
    strcat(filename, "/helpfile");
    helpfile.open(filename, ios::ate | ios::in | ios::nocreate);
    if (!helpfile) {
	helpfile.clear();
	int x_pos, y_pos;
	getsyx(y_pos, x_pos);
	wmove(win, 0, 0);
	wnoutrefresh(win);
	wclrtobot(win);
	wnoutrefresh(win);
	setsyx(y_pos, x_pos);
	doupdate();
	write("Can't open or read helpfile\nPress F1 again to exit");
	winrefresh();
    }
    else {
	streampos size = helpfile.tellg();
	if (!(buffer_ptr = new Display_buffer_with_curses((int)size, FALSE, ncols, nrows - 9))) {
	    cerr << "Memory allocation error in HelpWin::HelpWin()" << endl;
	    exit(MEM_ERROR);
	}
	helpfile.seekg(0); // go to beginning of file
	char letter;
	while (helpfile.get(letter)) store_letter(letter);
	helpfile.close();
	wmove(win, 0, 0);
	clear_win();
	buffer_ptr->display_first_page(win);
    }
}


ReceiveWin::ReceiveWin(Tnc& tnc_): tnc(tnc_), helpwin_ptr(0), helpwin_flag(FALSE) {
    receivewindow_rows = nrows - 9;
    win = newwin(receivewindow_rows, ncols, 8, 0);
    scrollok(win, TRUE);
    set_cursor_status(off);
    if (have_colour) wbkgd(win, COLOR_PAIR(1));
    show();
    int port;
    int stream;
    for (port = 0; port < 2; port++) {
        for (stream = 0; stream < MAXUSERS; stream++) {
	    if (!(buffer_ptr[stream][port] = new Display_buffer_with_curses(5000, TRUE,
		   ncols, receivewindow_rows))) {
	        cerr << "Memory allocation error in ReceiveWin::ReceiveWin()" << endl;
		exit(MEM_ERROR);
	    }
	}
    }
}

ReceiveWin::~ReceiveWin(void) {
    int port;
    int stream;
    for (port = 0; port < 2; port++) {
        for (stream = 0; stream < MAXUSERS; stream++) {
	    delete buffer_ptr[stream][port];
	}
    }
    delwin(win);
}

const void* ReceiveWin::is_scrolling(void) {
    if (helpwin_flag) return (const void*) 1;  // if we are displaying a helpfile, pretend we are scrolling
    else return buffer_ptr[tnc.tnc_func.active_stream()][tnc.tnc_func.active_port]->is_scrolling();
}

void ReceiveWin::display_helpfile(void) {
    helpwin_ptr = new HelpWin;
    if (!helpwin_ptr) {
        cerr << "Memory allocation error in ReceiveWin::display_helpfile()" << endl;
	exit(MEM_ERROR);
    }
    helpwin_flag = TRUE;
}

void ReceiveWin::write(char letter) {
    if (!is_scrolling()) {
        waddch(win, (uchar)letter);
	if (cursor_status == on) {
	    wnoutrefresh(win);
	    get_cur_pos();
	}
	refresh_winlist(active_popups);
    }
}

void ReceiveWin::write(char letter, int rowpos, int colpos) {
    if (!is_scrolling()) {
        wmove(win, rowpos, colpos);
	waddch(win, (uchar)letter);
	if (cursor_status == on) {
	    wnoutrefresh(win);
	    get_cur_pos();
	}
	refresh_winlist(active_popups);
    }
}

void ReceiveWin::write(const char* message) {
    if (!is_scrolling()) {
        waddstr(win, message);
	if (cursor_status == on) {
	    wnoutrefresh(win);
	    get_cur_pos();
	}
	refresh_winlist(active_popups);
    }
}

void ReceiveWin::write(const char* message, int rowpos, int colpos) {
    if (!is_scrolling()) {
        wmove(win, rowpos, colpos);
	waddstr(win, message);
	if (cursor_status == on) {
	    wnoutrefresh(win);
	    get_cur_pos();
	}
	refresh_winlist(active_popups);
    }
}

SendWin::SendWin(void) {
    win = newwin(5, ncols, 2, 0);
    scrollok(win, TRUE);
    set_cursor_status(on);
    if (have_colour) wbkgd(win, COLOR_PAIR(1));
    wmove(win, 0, 0);   // set up cursor
    wrefresh(win);
    get_cur_pos();
}

PopupWin::PopupWin(int n_rows_, int n_cols_, int row_pos, int col_pos):
                     popup_registered(FALSE), popup_rows(n_rows_), popup_cols(n_cols_) {
    win = newwin(n_rows_, n_cols_, row_pos, col_pos);
    scrollok(win, TRUE);
    set_cursor_status(off);
    if (have_colour) wbkgd(win, COLOR_PAIR(1));
    wmove(win, 0, 0);
    wattron(win, A_REVERSE);
    const int MAX = n_rows_ * n_cols_;
    int count;
    for (count = 0; count < MAX; count++) waddch(win, ' ');
    wmove(win, 0, 0);
}

void PopupWin::show(void) {
    if (!popup_registered) {
  	active_popups.add(this);
  	popup_registered = TRUE;
    }
    touchwin(win);
    if (cursor_status == off) {
        wnoutrefresh(win);
	set_cur_pos();
	doupdate();
    }
    else {
// if switching beween windows, both of which show cursors when in foreground, we want to reseat the cursor
        wrefresh(win);
	get_cur_pos();
    }

}

void PopupWin::unshow(void) {
    if (popup_registered) {
        set_cursor_status(off);
	wmove(win, 0, 0);               // we have to clear the screen with spaces rather
	const int MAX = popup_rows * popup_cols;// using werase() in order to retain the reverse
	int count;                              // attribute for text
	for (count = 0; count < MAX; count++) waddch(win, ' ');
	winrefresh();
// find the current window in active_popups in order to delete it in the list
	active_popups.reset(WinList::top_end);
	const Screen* temp;
	while ((temp = active_popups.inspect(WinList::down)) != 0 && temp != (Screen*)this);
	if (temp) active_popups.extract();
	popup_registered = FALSE;
    }
}

CopyrightScreen::CopyrightScreen(void) {
    win = newwin(nrows, ncols, 0, 0);
    set_cursor_status(off);
    scrollok(win, TRUE);

// const char copyright_msg[] is defined in gpl.h

    if (!(buffer_ptr = new Display_buffer_with_curses(strlen(copyright_msg), FALSE, ncols, nrows))) {
        cerr << "Memory allocation error in CopyrightScreen::CopyrightScreen()" << endl;
	exit(MEM_ERROR);
    }

    wmove(win, 0, 0);
    char* letter_p;
    for (letter_p = (char*)copyright_msg; *letter_p; letter_p++) store_letter(*letter_p);
    clear_win();
    buffer_ptr->display_first_page(win);
}

void WinList::add(Screen* screen) {
    WinNode* node = new WinNode(screen, top);
    if (!node) {
        cerr << "Memory allocation error in WinList::add()" << endl;
	exit(MEM_ERROR);
    }
    if (!bottom) bottom = node; // if this is the first entry, attach it to bottom
    if (top) top->next = node; // set WinNode::next in previous entry to new entry
    top = node; // now update top
}

const Screen* WinList::inspect(Direction dir) {
    Screen* temp;
    if (inspect_pos) { // not at end of inspection
        extract_pos = inspect_pos;
        temp = inspect_pos->screen; // this is the return value
	if (dir == down) inspect_pos = inspect_pos->prev; // set inspect_pos to previous node in the list if going down the list
	else inspect_pos = inspect_pos->next; // else we are going up the list
    }
    else temp = 0;
    return temp;
}

void WinList::extract(void) {
    if (extract_pos) {
        if (top == bottom) top = bottom = 0; // only one entry in WinList object
	else {
	// if there is a higher item, then reset the link
	    if (extract_pos->next) extract_pos->next->prev = extract_pos->prev;
	// if not, the item to be extracted must be at the top, so reset top
	    else top = extract_pos->prev; // the extracted item is at the top, so reset top
	// if there is a lower item, then reset the link
	    if (extract_pos->prev) extract_pos->prev->next = extract_pos->next;
	// if not, the item to be extracted must be at the bottom, so reset bottom
	    else bottom = extract_pos->next;
	}
	delete extract_pos;
	extract_pos = 0;
    }
}

void WinList::clear_list(void) {
    WinNode* temp;
    reset(top_end);
    while(inspect_pos) {
        temp = inspect_pos;
	inspect(down);
        delete temp;
    }
    top = bottom = 0;
}

void refresh_winlist(WinList& form) {
    form.reset(WinList::bottom_end);
    Screen* screen_p;
    while((screen_p = (Screen*)form.inspect(WinList::up)) != 0) screen_p->show();
}
